# -*- coding: utf-8 -*-

"""
    tickets.showes
    ~~~~~~~~~~~~~~

    Show tickets model.
"""

import os
import re
import sys
from .utils import colored

import requests
from bs4 import BeautifulSoup
from prettytable import PrettyTable
from datetime import datetime, timedelta
from requests.exceptions import ConnectionError


__all__ = ['ShowTicketsQuery']


SHOWES_QUERY_URL = 'http://www.damai.cn/projectlist.do'

# ERROR MSG
QUERY_DAYS_INVALID = 'Invalid days.'
CITY_NOT_FOUND = 'City not found.'
SHOW_NOT_FOUND = 'No result.'
NETWORK_CONNECTION_FAIL = 'Network connection failed.'


class ShowesCollection(object):

    """A set of showes from a query."""

    #: The header of every column
    headers = '主题 时间 场馆 票价'.split()

    def __init__(self, rows):
        self._rows = rows

    def __repr__(self):
        return '<ShowTicketsCollection size={}>'.format(len(self))

    def __len__(self):
        return len(self._rows)

    def pretty_print(self):
        pt = PrettyTable()
        if len(self) == 0:
            pt._set_field_names(['Sorry,'])
            pt.add_row([SHOW_NOT_FOUND])
        else:
            pt._set_field_names(self.headers)
        for row in self._rows:
            pt.add_row(row)
        print(pt)


class ShowTicketsQuery(object):

    """Perform a show query."""

    #: show query params
    show_types = {
        '演唱会': {'mcid': 1, 'ccid': ''},
        '音乐会': {'mcid': 2, 'ccid': ''},
        '话剧': {'mcid': 3, 'ccid': 19},
        '歌剧': {'mcid': 3, 'ccid': 20},
        '音乐剧': {'mcid': 3, 'ccid': 22},
        '儿童剧': {'mcid': 3, 'ccid': 23},
        '歌舞剧': {'mcid': 3, 'ccid': 21},
        '舞蹈': {'mcid': 4, 'ccid': ''},
        '相声': {'mcid': 5, 'ccid': 27},
        '魔术': {'mcid': 5, 'ccid': 28},
        '马戏': {'mcid': 5, 'ccid': 29},
        '杂技': {'mcid': 5, 'ccid': 30},
        '戏曲': {'mcid': 5, 'ccid': 31},
        '比赛': {'mcid': 6, 'ccid': ''}
    }

    def __init__(self, city, show_type, days=15):
        self.city = city
        self.show_type = show_type
        self.days = days

    def __repr__(self):
        return '<ShowTicketsQuery city={} types={} days={}>'.format(
            self._city, self._type, self._days
        )

    @property
    def cities(self):
        filepath = os.path.join(
            os.path.dirname(__file__),
            'datas', 'cities.dat'
        )
        d = {}
        with open(filepath, 'r') as f:
            for line in f.readlines():
                name, number = line.split()
                d.setdefault(name, int(number))
        return d

    @property
    def _city_id(self):
        ci = self.cities.get(self.city)
        if not ci:
            exit_after_echo(CITY_NOT_FOUND)
        return ci

    @property
    def _show_type(self):
        st = self.show_types.get(self.show_type)
        if not st:
            exit_after_echo(SHOW_NOT_FOUND)
        return st

    @property
    def date_range(self):
        """Generate date range according to the `days` user input."""
        try:
            days = int(self.days)
        except ValueError:
            exit_after_echo(QUERY_DAYS_INVALID)

        if days < 1:
            exit_after_echo(QUERY_DAYS_INVALID)
        start = datetime.today()
        end = start + timedelta(days=days)
        return (
            datetime.strftime(start, '%Y-%m-%d'),
            datetime.strftime(end, '%Y-%m-%d')
        )

    def _build_params(self):
        start, end = self.date_range
        return dict(
            cityID=self._city_id, isText=1,
            pageIndex=1, startDate=start,
            endDate=end, order=2,
            **self._show_type
        )

    def parse(self, items):
        """Parse `主题`, `时间`, `场馆`, 票价` in every item."""
        rows = []
        for i, item in enumerate(items):
            color = colored.green if i & 1 else colored.red
            theme = color(item.find(class_='ico').a.text.strip())
            text = item.find(class_='mt10').text.strip()
            mix = re.sub('\s+', ' ', text).split('：')
            time = mix[1][:-3]
            place = color(mix[2][:-7])
            price = item.find(class_='price-sort').text.strip()
            rows.append([theme, time, place, price])
        return rows

    def query(self):
        params = self._build_params()
        rows = []
        while True:
            try:
                r = requests.get(SHOWES_QUERY_URL, params=params)
            except ConnectionError:
                exit_after_echo(NETWORK_CONNECTION_FAIL)
            soup = BeautifulSoup(r.text, 'html.parser')
            items = soup.find_all(class_='ri-infos')
            if not items:
                return ShowesCollection(rows)
            rows += self.parse(items)
            params['pageIndex'] += 1



# -*- coding: utf-8 -*-

"""
    tickets.core
    ~~~~~~~~~~~~

    tickets command-line interface.
"""

import sys
import requests
import colorama
from .utils import args, exit_after_echo
from .showes import ShowTicketsQuery
from .trains import TrainTicketsQuery


try:
    from requests.packages.urllib3.exceptions import (
            SNIMissingWarning,
            InsecureRequestWarning,
            InsecurePlatformWarning
    )

# Not show warings
    requests.packages.urllib3.disable_warnings(SNIMissingWarning)
    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
    requests.packages.urllib3.disable_warnings(InsecurePlatformWarning)
except ImportError:
    pass


def show_usage():
    """
Usage:
    tickets [-dgktz] <from> <to> <date>
    tickets <city> <show-type> <within-days>

You can go to `tickets -h` for more details.
    """
    pass


def cli():
    """Tickets query via the command line.

Usage:
    tickets [-dgktz] <from> <to> <date>
    tickets <city> <show> [<days>]

 Arguments:
    from             出发站
    to               到达站
    date             查询日期

    city             查询城市
    show             演出的类型
    days             查询近(几)天内的演出, 若省略, 默认15


Options:
    -h, --help       显示该帮助菜单.
    -d               动车
    -g               高铁
    -k               快速
    -t               特快
    -z               直达

Show:
    演唱会 音乐会 比赛 话剧 歌剧 舞蹈 戏曲 相声 音乐剧 歌舞剧 儿童剧 杂技 马戏 魔术

Examples:
    tickets 南京 北京 201671
    tickets -k  南京南 上海 2016-7-1
    tickets -dg 上海虹桥 北京西 2016/7/1

    tickets 上海 演唱会
    tickets 北京 比赛 7
    """

    colorama.init()

    if args.is_asking_for_help:
        exit_after_echo(cli.__doc__, color=None)

    elif args.is_querying_show:
        q = ShowTicketsQuery(*args.as_show_query_params)

    elif args.is_querying_train:
        q = TrainTicketsQuery(*args.as_train_query_params)

    else:
        exit_after_echo(show_usage.__doc__, color=None)

    q.query().pretty_print()



# -*- coding: utf-8 -*-
"""
    tickets._compat
    ~~~~~~~~~~~~~~~

    Compatibility for Python2 and Python3.
"""
import sys


_ver = sys.version_info

#: Py2.*?
is_py2 = (_ver[0] == 2)

#: Py3.*?
is_py3 = (_ver[0] == 3)

if is_py2:
    unicode_type = unicode
    bytes_type = str
elif is_py3:
    unicode_type = str
    bytes_type = bytes


def to_unicode(value, encoding='utf-8'):
    if isinstance(value, unicode_type):
        return value

    if isinstance(value, bytes_type):
        return unicode_type(value, encoding=encoding)

    if isinstance(value, int):
        return unicode_type(str(value))

    return value



# -*- coding: utf-8 -*-

"""
    tickets.utils
    ~~~~~~~~~~~~~

    Some tool.
"""

import sys
import colorama


__all__ = ['args', 'colored', 'exit_after_echo', 'is_show_type']


def exit_after_echo(msg, color='red'):
    if color == 'red':
        print(colored.red(msg))
    else:
        print(msg)
    exit()

show_types_list = (
    '演唱会 音乐会 比赛 话剧 歌剧 戏曲 相声 ' +
    '音乐剧 歌舞剧 儿童剧 舞蹈 杂技 马戏 魔术'
).split()


is_show_type = frozenset(show_types_list).__contains__


class Args(object):

    """A simple customed args parser for `tickets`."""

    def __init__(self, args=None):
        self._args = sys.argv[1:]

    def __repr__(self):
        return '<args {}>'.format(repr(self._args))

    def __len__(self):
        return len(self._args)

    @property
    def all(self):
        return self._args

    def get(self, idx):
        try:
            return self.all[idx]
        except IndexError:
            return None

    @property
    def is_null(self):
        return len(self.all) == 0

    @property
    def options(self):
        """Train tickets query options."""
        arg = self.get(0)
        if arg.startswith('-') and not self.is_asking_for_help:
            return arg[1:]
        return ''.join(x for x in arg if x in 'dgktz')

    def contain_show_type(self):
        arg = self.get(2)
        if is_show_type(arg):
            return arg
        return None

    @property
    def is_asking_for_help(self):
        arg = self.get(0)
        if arg in ('-h', '--help'):
            return True
        return False

    @property
    def is_querying_show(self):
        arg = self.get(1)
        if len(self.all) not in (2, 3):
            return False
        if is_show_type(arg):
            return True
        return False

    @property
    def is_querying_train(self):
        if len(self.all) not in (3, 4):
            return False
        if self.is_querying_show:
            return False
        return True

    @property
    def as_train_query_params(self):
        opts = self.options
        if opts:
            # apped valid options to end of list
            return self.all[1:] + [opts]
        return self.all

    @property
    def as_show_query_params(self):
        return self.all


class Colored(object):

    """A simple wrapper based on colorama.

    Fore: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
    Back: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
    Style: DIM, NORMAL, BRIGHT, RESET_ALL

    """

    def color_str(self, color, s, style='NORMAL'):
        return '{}{}{}{}'.format(
            getattr(colorama.Fore, color),
            getattr(colorama.Style, style),
            s,
            colorama.Style.RESET_ALL
        )

    def red(self, s):
        return self.color_str('RED', s)

    def green(self, s):
        return self.color_str('GREEN', s)


args = Args()
colored = Colored()



# -*- coding: utf-8 -*-

"""
    tickets.trains
    ~~~~~~~~~~~~~~

    Train tickets model.
"""

import os
import re
import sys
import json
from datetime import datetime
from collections import OrderedDict
from .utils import colored, exit_after_echo

import requests
from prettytable import PrettyTable
from requests.exceptions import ConnectionError


__all__ = ['TrainTicketsQuery']


QUERY_URL = 'https://kyfw.12306.cn/otn/lcxxcx/query'

# ERROR MSG
FROM_STATION_NOT_FOUND = 'From station not found.'
TO_STATION_NOT_FOUND = 'To station not found.'
INVALID_DATE = 'Invalid query date.'
NETWORK_CONNECTION_FAIL = 'Network connection failed.'


class TrainsCollection(object):

    """A set of raw datas from a query."""

    headers = '车次 车站 时间 历时 商务 一等 二等 软卧 硬卧 软座 硬座 无座'.split()

    def __init__(self, rows, opts):
        self._rows = rows
        self._opts = opts

    def __repr__(self):
        return '<TrainsCollection size={}>'.format(len(self))

    def __iter__(self):
        i = 0
        while True:
            if i < len(self):
                yield self[i]
            else:
                yield next(self)
            i += 1

    def __len__(self):
        return len(self._rows)

    def _get_duration(self, row):
        duration = row.get('lishi').replace(':', '小时') + '分钟'
        # take 0 hour , only show minites
        if duration.startswith('00'):
            return duration[4:]
        # take <10 hours, show 1 bit
        if duration.startswith('0'):
            return duration[1:]
        return duration

    @property
    def trains(self):
        """Filter rows according to `headers`"""
        for row in self._rows:
            train_no = row.get('station_train_code')
            initial = train_no[0].lower()
            if not self._opts or initial in self._opts:
                train = [
                    # Column: '车次'
                    train_no,
                    # Column: '车站'
                    ''.join([
                        colored.red(row.get('from_station_name')),
                        '\n',
                        colored.green(row.get('to_station_name')),
                    ]),
                    # Column: '时间'
                    ''.join([
                        colored.red(row.get('start_time')),
                        '\n',
                        colored.green(row.get('arrive_time')),
                    ]),
                    # Column: '历时'
                    self._get_duration(row),
                    # Column: '商务'
                    row.get('swz_num'),
                    # Column: '一等'
                    row.get('zy_num'),
                    # Column: '二等'
                    row.get('ze_num'),
                    # Column: '软卧'
                    row.get('rw_num'),
                    # Column: '硬卧'
                    row.get('yw_num'),
                    # Column: '软座'
                    row.get('rz_num'),
                    # Column: '硬座'
                    row.get('yz_num'),
                    # Column: '无座'
                    row.get('wz_num')
                ]
                yield train

    def pretty_print(self):
        """Use `PrettyTable` to perform formatted outprint."""
        pt = PrettyTable()
        if len(self) == 0:
            pt._set_field_names(['Sorry,'])
            pt.add_row([TRAIN_NOT_FOUND])
        else:
            pt._set_field_names(self.headers)
            for train in self.trains:
                pt.add_row(train)
        print(pt)


class TrainTicketsQuery(object):

    """Docstring for TrainTicketsCollection. """

    def __init__(self, from_station, to_station, date, opts=None):

        self.from_station = from_station
        self.to_station = to_station
        self.date = date
        self.opts = opts

    def __repr__(self):
        return 'TrainTicketsQuery from={} to={} date={}'.format(
            self._from_station, self._to_station, self._date
        )

    @property
    def stations(self):
        filepath = os.path.join(
            os.path.dirname(__file__),
            'datas', 'stations.dat'
        )
        d = {}
        with open(filepath, 'r') as f:
            for line in f.readlines():
                name, telecode = line.split()
                d.setdefault(name, telecode)
        return d

    @property
    def _from_station_telecode(self):
        code = self.stations.get(self.from_station)
        if not code:
            exit_after_echo(FROM_STATION_NOT_FOUND)
        return code

    @property
    def _to_station_telecode(self):
        code = self.stations.get(self.to_station)
        if not code:
            exit_after_echo(TO_STATION_NOT_FOUND)
        return code

    @property
    def _valid_date(self):
        date = re.sub(r'[-/\:,]+', '', self.date)
        try:
            date = datetime.strptime(self.date, '%Y%m%d')
        except ValueError:
            exit_after_echo(INVALID_DATE)
        diff = date - datetime.today()
        if diff.days not in range(-1, 50):
            exit_after_echo(INVALID_DATE)
        return datetime.strftime(date, '%Y-%m-%d')

    def _build_params(self):
        """Have no idea why wrong params order can't get data.
        So, use `OrderedDict` here.
        """
        d = OrderedDict()
        d['purpose_codes'] = 'ADULT'
        d['queryDate'] = self._valid_date
        d['from_station'] = self._from_station_telecode
        d['to_station'] = self._to_station_telecode
        return d

    def query(self):

        params = self._build_params()

        try:
            r = requests.get(QUERY_URL, params=params, verify=False)
        except ConnectionError:
            exit_after_echo(NETWORK_CONNECTION_FAIL)

        try:
            rows = r.json()['data']['datas']
        except KeyError:
            rows = []

        return TrainsCollection(rows, self.opts)



#!/usr/bin/env python3

from codecs import open
from tickets import __version__
from setuptools import setup, find_packages


def read(f):
    return open(f, encoding='utf-8').read()


setup(
    name='tickets',
    version=__version__,
    description='Train tickets query via command line.',
    long_description=read('README.rst') + '\n\n' + read('HISTORY.rst'),
    author='protream',
    author_email='protream@gmail.com',
    url='https://github.com/protream/tickets',
    packages=[
        'tickets'
    ],
    py_modules=['run'],
    include_package_data=True,
    platforms='any',
    install_requires=[
        'prettytable',
        'requests',
        'bs4'
    ],
    entry_points={
        'console_scripts': ['tickets=run:cli']
    },
    license='MIT',
    zip_safe=False,
    classifiers=[
         'Environment :: Console',
         'Programming Language :: Python',
         'Programming Language :: Python :: 3',
         'Programming Language :: Python :: 3.1',
         'Programming Language :: Python :: 3.2',
         'Programming Language :: Python :: 3.3',
         'Programming Language :: Python :: 3.4',
         'Programming Language :: Python :: 3.5',
         'Programming Language :: Python :: Implementation :: CPython'
    ]
)
